/* $Id: i18n.c,v 1.7 1999/03/09 19:20:11 frog Exp $ */

#include <stdio.h>
#include <stdlib.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/Xresource.h>
#include <string.h>

/*
 * Xlocale.h includes <locale.h> or the non-standard X substitutes
 * depending on the X_LOCALE compilation flag.
 */
#include <X11/Xlocale.h>

#define DEFAULT_FONT	"-misc-gothic-medium-r-normal--16-*-*-*-c-*-*"

/* For error messages */
char *program_name ;
char* class_name ;
/*
 * Whether Xlib performs directional dependent drawing or not.
 * Assume Xlib does not perform as default.
 */
Bool is_implicit_ddd = False;
/*
 * Whether Xlib performs contextual dependent drawing or not.
 * Assume Xlib does not perform as default.
 */
Bool is_implicit_cdd = False;
/* Default Orientation */
XOMOrientation *orientation ;

/* Private convenient functions */
static XOM	open_om(Display *, char *, char *);
static XOC 	create_oc(XOM, char *, char *);
static XIC      create_ic( XIM, Window, XrmDatabase, char *, char *);

static XIMStyle DetermineStyle(XIM, XrmDatabase, char *, char *,
                               XIMStyle *);
static void	GetPreferredGeometry( XIC, char *, XRectangle *);
static void	SetGeometry( XIC, char *, XRectangle *);

static void 	app_im_init( Display *);
static void     im_instantiate_cb(Display *, XPointer, XPointer);

/* global variables */
XOM om;
XOC oc;
XIM im;
XIC ic;
int screen;
Window win;

int main(int argc, char *argv[])
{
  Display *display ;
  GC gc;
  XEvent event;
  XIMStyle style;
  XRectangle preedit_area;
  XRectangle status_area;
  char **missing_charsets;
  int num_missing_charsets = 0;
  char *default_string;
  char string[200];
  int str_len = 0;
  int i;
  XWMHints hints;
  unsigned long   mask;
  XrmDatabase	rdb;

  program_name = strdup(argv[0]);
  class_name = strdup(argv[0]);

  /*
   * The error messages in this program are all in English.
   * In a truly internationalized program, they would not
   * be hardcoded; they would be looked up in a database of
   * some sort.
   */
  if (setlocale(LC_ALL, "") == NULL) {
    fprintf(stderr, "%s: cannot set locale.\n",program_name);
    exit(1);
  }
  if ((display = XOpenDisplay(NULL)) == NULL) {
    fprintf(stderr, "%s: cannot open Display.\n", program_name);
    exit(1);
  }
  if (!XSupportsLocale()) {
    fprintf(stderr, "%s: X does not support locale \"%s\".\n",
	program_name, setlocale(LC_ALL, NULL));
    exit(1);
  }
  if (XSetLocaleModifiers("") == NULL) {
    fprintf(stderr, "%s: Warning: cannot set locale modifiers.\n",
	program_name);

    /* Try setting the local IM as fallback */
    if (XSetLocaleModifiers("@im=none") == NULL) {
      fprintf(stderr, "%s: Warning: cannot set fallback locale modifiers.\n",
	  program_name);
    }
  }
  /* Opening an output method */
  if((om = open_om(display, program_name, class_name)) == NULL){
    fprintf(stderr, "%s: Cannot open XOM for the locale \"%s\".\n",
	program_name, setlocale(LC_ALL, NULL));
    exit(1);
  }
  /* Creating an output context */
  if((oc = create_oc(om, program_name, class_name)) == NULL){
    fprintf(stderr, "%s: Cannot create XOC for the locale \"%s\".\n",
	program_name, setlocale(LC_ALL, NULL));
    exit(1);
  }

  screen = DefaultScreen(display);
  win = XCreateSimpleWindow(display, RootWindow(display, screen), 0, 0,
                            400, 100, 2, BlackPixel(display,screen),
                            WhitePixel(display,screen));
  XStoreName(display, win, "XIM Demo - Dynamic Connection");

  gc = DefaultGC(display, screen);
  XSetForeground(display,gc,BlackPixel(display,screen));
  XSetBackground(display,gc,WhitePixel(display,screen));
  hints.flags = InputHint;
  hints.input = True;
  XSetWMHints(display, win, &hints);
  mask = KeyPressMask | StructureNotifyMask | FocusChangeMask | ExposureMask;
  XSelectInput(display, win, mask);
  XMapRaised(display, win);
  XFlush(display);

  /*
   * Register XIM instantiate callback if IM Server is not running,
   * otherwise call application's IM initialize routine
   */
  rdb = XrmGetDatabase(display);
  XRegisterIMInstantiateCallback(display, rdb,
                                 program_name, class_name,
                                 im_instantiate_cb, NULL );
  while(1) {
    int         buf_len = 255;
    char        *buffer = (char *)malloc(buf_len);
    int         len;
    KeySym	keysym;
    Status      status;
    Bool        redraw = False;

    XNextEvent(display, &event);
    if( XFilterEvent(&event, None) )
      continue;
    switch (event.type) {
      case Expose:
        /* draw the string at a hard-coded location */
        if (event.xexpose.count == 0)
          XmbDrawString(display, win, (XFontSet)oc, gc, 10, 50,
                        string, str_len);
        break;
      case KeyPress:
        if( ic ){
          len = XmbLookupString(ic, (XKeyPressedEvent*)&event, buffer,
                                buf_len, &keysym, &status);
          if (status == XBufferOverflow) {
            buf_len = len;
            buffer = (char *)realloc((char *)buffer,
                                     buf_len * sizeof(wchar_t));
            len = XmbLookupString(ic, (XKeyPressedEvent*)&event,
                                  buffer,buf_len, &keysym, &status);
          }

          redraw = False;
          switch (status) {
            case XLookupNone:
              break;
            case XLookupKeySym:
            case XLookupBoth:
              /* Handle backspacing, and <Return> to exit */
              if ((keysym == XK_Delete) || (keysym == XK_BackSpace)) {
                if( str_len > 0 )
		  str_len--;
                redraw = True;
                break;
              }
              if (keysym == XK_Escape )
		exit(0);
              if (status == XLookupKeySym)
		break;
            case XLookupChars:
              for(i=0; i < len; i++)
                string[str_len++] = buffer[i];
              redraw = True;
              break;
          }
        }
	else {
          /* Until ic is created, behave as C locale client. */
          len = XLookupString((XKeyEvent *)&event, buffer, buf_len,
                              &keysym, NULL);
          for(i = 0; i < len; i++ )
            string[str_len++] = buffer[i];
          string[str_len] = 0;
          redraw = (len > 0);
        }
        /* do a very simple-minded redraw, if needed */
        if( redraw ) {
          XClearWindow(display, win);
          XmbDrawString(display, win, (XFontSet)oc, gc, 10, 50, string,
                        str_len);
        }
        break;
      case ConfigureNotify:
        /*
         * When the window is resized, we should re-negotiate the
         * geometry of the Preedit and Status area, if they are used
         * in the interaction style.
         */
        if( ic )
        {
          XGetICValues(ic, XNInputStyle, &style, NULL);
          if( style & XIMPreeditArea ) {
            preedit_area.width = event.xconfigure.width*4/5;
            preedit_area.height = 0;
            GetPreferredGeometry(ic, XNPreeditAttributes,
                                 &preedit_area);
            preedit_area.x = event.xconfigure.width -
                             preedit_area.width;
            preedit_area.y = event.xconfigure.height -
                             preedit_area.height;
            SetGeometry(ic, XNPreeditAttributes, &preedit_area);
          }
          if( style & XIMStatusArea ) {
            status_area.width = event.xconfigure.width/5;
            status_area.height = 0;
            GetPreferredGeometry(ic, XNStatusAttributes,
                                 &status_area);
            status_area.x = 0;
            status_area.y = event.xconfigure.height -
                            status_area.height;
            SetGeometry(ic, XNStatusAttributes, &status_area);
          }
        }
        break;
      case FocusIn:
        if (event.xany.window == win) {
          if(ic)
            XSetICFocus(ic);
        }
        break;
      case FocusOut:
        if (event.xany.window == win) {
          if(ic)
            XUnsetICFocus(ic);
        }
        break;
    }
  }
}

XOM open_om(Display *display, char *ProgramName, char *ClassName)
{
  XOM om ;
  XrmDatabase	rdb = XrmGetDatabase(display);

  if((om = XOpenOM(display, rdb, ProgramName, ClassName)) == NULL){
    fprintf(stderr, "%s: Cannot open Output Method\n", program_name);
    return (XOM)NULL ;
  }
  if(XGetOMValues(om, XNOrientation, &orientation, NULL)){
    /* Assume Left to Right, Top to Bottom as fallback. */
    orientation = (XOMOrientation *)malloc(sizeof(XOMOrientation));
    /*
     * As this book goes to press, there are bugs in Xlib.h.
     * The specification defines num_orientation and orientation
     * as members of XOMOrientation, but Xlib.h defines them
     * as num_orient and orient
     */
    orientation->num_orientation = 1 ;
    orientation->orientation = (XOrientation *)malloc(sizeof
                               (XOrientation));
    *orientation->orientation = XOMOrientation_LTR_TTB ;
  }
  XGetOMValues(om, XNDirectionalDependentDrawing, &is_implicit_ddd, NULL);
  XGetOMValues(om, XNContextualDrawing, &is_implicit_cdd, NULL);

  return om ;
}

XOC create_oc(XOM om, char *ProgramName, char *ClassName)
{
  int	i ;
  char	*basename = DEFAULT_FONT;
  XOC	oc;
  XOMCharSetList list;

  if( (oc = XCreateOC(om, XNBaseFontName, basename, NULL)) == NULL)
  {
    fprintf(stderr, "%s: Cannot create Output Context\n", program_name);
    return (XOC)NULL ;
  }

  if(!XGetOCValues(oc, XNMissingCharSet, &list, NULL)){
    if( list.charset_count > 0){
      fprintf(stderr, "%s: Warning: the following charsets are missing\n",
	  program_name);
      for(i=0; i < list.charset_count ; i++){
        fprintf(stderr, "%s: %s\n", program_name, list.charset_list[i]);
      }
    }
  }
  return oc;
}

/*
 * Callback to instantiate an input method on demand.
 * invoke input method related application initialization routine.
 */
static void im_instantiate_cb(Display *display, XPointer client, XPointer call)
{
  app_im_init(display);
#if 0
  if( fail )
    XUnregisterIMInstantiateCallback (display, NULL, NULL, NULL,
                                      im_instantiate_cb, NULL);
#endif
}

static void destroy_callback(XIM im, XPointer client, XPointer call)
{
  im = NULL;
  ic = NULL;
  XUnregisterIMInstantiateCallback (display, NULL, NULL, NULL,
                                    im_instantiate_cb, NULL);
}

static is_supported(char *val, XIMValuesList *val_list)
{
  unsigned short i ;
  for(i = 0 ; i < val_list->count_values ; i++){
    /* Workarround: double check */
    if( !val_list->supported_values )
      return False ;
    if(!strcmp(val, val_list->supported_values[i])){
      return True;
    }
  }
  return False;
}

/*
 * Application specific XIM initialization
 */
void app_im_init(Display *display)
{
  XIMCallback	destroy;
  unsigned long mask;
  /* Supported im/ic values list. */
  XIMValuesList im_values_list ;
  XIMValuesList ic_values_list ;
  XrmDatabase	rdb = XrmGetDatabase(display);

  if(!(im = XOpenIM(display, rdb, program_name, class_name))){
    return;
  }

  /*
   * Retrieve supported XIM velues list and XIC values list
   * for later usage.
   */
  if(XGetIMValues(im, XNQueryIMValuesList, &im_values_list, NULL)){
    fprintf(stderr, "%s: Warning: XNQueryIMValuesList failed\n",
            program_name);
  }

  /*
   * Check XNDestroyCallback is supported or not by using the 
   * just retrieved im values list
   */
  if(is_supported(XNDestroyCallback, &im_values_list)){
    destroy.callback = destroy_callback;
    destroy.client_data = NULL;
    XSetIMValues(im, XNDestroyCallback, &destroy, NULL );
  } else {
    fprintf(stderr,
	"%s: Warning: Xlib does not support destroy callback\n",
	program_name);
  }
  if(is_supported(XNQueryICValuesList, &im_values_list)){
    /*
     * Retrieve supported XIC velues list and XIC values list
     * for later usage.
     */
    if(XGetIMValues(im, XNQueryICValuesList, &ic_values_list, NULL)){
      fprintf(stderr, "%s: Warning: XNQueryICValuesList failed\n",
              program_name);
    }
  }
  /*
   * Creating an input context
   */
  if((ic = create_ic(im, win, rdb, program_name, class_name)) == NULL){
    fprintf(stderr, "%s: Cannot create XIC for the locale \"%s\".\n",
	program_name, setlocale(LC_ALL, NULL));
    exit(1);
  }
  XGetICValues(ic, XNFilterEvents, &mask, NULL);
  mask |= (ExposureMask | KeyPressMask | FocusChangeMask);
  XSelectInput(display, win, mask);
}

/*
 * Create an IC with best input style in given environment.
 */
XIC create_ic( XIM im, Window client_window, XrmDatabase rdb,
               char *res_name, char	*res_class )
{
  XIMStyles supported_list ;
  XIC ic ;
  static XIMStyle style ;
  /*
   * The input styles supported by this application.
   */
  static XIMStyle supported_styles[] = {
#ifdef CALLBACK_SUPPORT
    XIMPreeditCallbacks|XIMStatusArea,
    XIMPreeditCallbacks|XIMStatusNothing,
    XIMPreeditCallbacks|XIMStatusNone,
    XIMPreeditCallbacks|XIMStatusCallbacks ,
#endif
    XIMPreeditPosition|XIMStatusArea,
    XIMPreeditPosition|XIMStatusNothing,
    XIMPreeditPosition|XIMStatusNone,
#ifdef CALLBACK_SUPPORT
    XIMPreeditPosition|XIMStatusCallbacks,
#endif
    XIMPreeditArea|XIMStatusArea,
    XIMPreeditArea|XIMStatusNothing,
    XIMPreeditArea|XIMStatusNone,
#ifdef CALLBACK_SUPPORT
    XIMPreeditArea|XIMStatusCallbacks,
#endif
    XIMPreeditNothing|XIMStatusArea,
    XIMPreeditNothing|XIMStatusNothing,
    XIMPreeditNothing|XIMStatusNone,
#ifdef CALLBACK_SUPPORT
    XIMPreeditNothing|XIMStatusCallbacks,
#endif
    XIMPreeditNone|XIMStatusArea,
    XIMPreeditNone|XIMStatusNothing,
    XIMPreeditNone|XIMStatusNone,
#ifdef CALLBACK_SUPPORT
    XIMPreeditNone|XIMStatusCallbacks,
#endif
    0
    };

  if(!style){
    style = DetermineStyle(im, rdb, res_name, res_class,
                           supported_styles);
  }
  /* for other style, we must supply extra information */
  style = XIMPreeditNothing | XIMStatusNothing;
  if((ic = XCreateIC(im,
                     XNInputStyle, style,
                     XNClientWindow, client_window,
                     XNFocusWindow, client_window,
                     NULL)) == NULL) {
    fprintf(stderr, "%s: Cannot create an IC\n", program_name);
  }
  return(ic);
}

/**
 * DetermineStyle() returns a best input style from 20 possible 
 * input styles, by taking user's preference priority, system default
 * priority per locale if exist and the suppoted styles which is queried 
 * by XGetIMValues(). 
 **/
#ifndef XLOCALEDIR
#define XLOCALEDIR "/usr/<acronym>X11R6</acronym>/lib/X11/locale"
#endif
#define STYLEFILEPATHSZMAX 30
/*
 * In order to cache the multiple best style per locale, per IM, per
 * individual resource name and class, this complicated structure
 * is needed.
 */
typedef struct _style_cache {
    struct _style_cache *next ;
    char	*locale ;
    XIM		im ;
    XrmDatabase rdb;
    char	*res_name ;
    char	*res_class ;
    XIMStyle    best_style ;
    XIMStyles	*user_styles ;
    XIMStyles	*im_dep_styles ;
    XIMStyles	*lc_dep_styles ;
} style_cache_t ;
/*
 * Fallback default
 */
#define DEFAULTStyle 	(XIMPreeditNothing|XIMStatusNothing)
static XIMStyles	*StrToIMStyles();
static XIMStyle		BestStyle();

static XIMStyle DetermineStyle(XIM im, XrmDatabase rdb,
    char *res_name, char *res_class, XIMStyle *supported_styles)
{
  static style_cache_t *p = 0 ;
  static XIMStyles *lc_dep_styles = 0 ;
  char *rmtype ;
  XrmValue rmval ;
  if(!p){
    if((p = (style_cache_t *)calloc(sizeof(style_cache_t),1))==NULL) {
      return(DEFAULTStyle);
    }
  } else {
    /* check the cache first */
    XIMStyles *lc_dep_styles = 0 ;
    char *locale = setlocale(LC_CTYPE, NULL);
    for(;;) {
      if(!strcmp(p->locale, locale)){
        if(p->im == im &&
            p->rdb == rdb &&
            !strcmp(p->res_name, res_name) &&
            !strcmp(p->res_class, res_class)) {
          return(p->best_style);
        }
        lc_dep_styles = p->lc_dep_styles ;
      }
      if (p->next){
        p = p->next ;
      } else {
        break ;
      }
    }
    if((p->next = (style_cache_t *)calloc(sizeof(style_cache_t),1))
        ==NULL) {
      return(DEFAULTStyle);
    }
    p = p->next ;
    if( lc_dep_styles )
      p->lc_dep_styles = lc_dep_styles ;
  }
  /* first time for this combination */
  p->locale = strdup(setlocale(LC_CTYPE, NULL));
  p->im = im ;
  p->rdb = rdb ;
  p->res_name = strdup(res_name);
  p->res_class = strdup(res_class);
  XGetIMValues(im, XNQueryInputStyle, &p->im_dep_styles, NULL);
  if( XrmGetResource(rdb, res_name, res_class, &rmtype, &rmval) )
  {
    /* We should use user's priority list */
    p->user_styles = StrToIMStyles(rmval.size, rmval.addr);
    p->best_style = BestStyle(p->user_styles, p->im_dep_styles,
	supported_styles) ;
  } else {
    /* We will use system's per-locale priority list */
    if(!p->lc_dep_styles){
#ifdef USESYSTEMDEFAULTSTYLE
      /* The code in this block runs once per locale. */
      int fd ;
      static int size = 0 ;
      static char *fn ;
      char *s ;
      static char *xlocalehome=0 ;

      if(!size){
        size = getpagesize();
      }
      if(!xlocalehome){
        if(!(xlocalehome = (char *)getenv("XLOCALEDIR"))){
          xlocalehome = XLOCALEDIR ;
        }
        fn = (char *)malloc(strlen(xlocalehome)+strlen(p->locale)
                            +STYLEFILEPATHSZMAX) ;
        sprintf(fn, "%s/%s/InputStyles", xlocalehome, p->locale);
      }
      if( (fd = open(fn, O_RDONLY)) == -1 ){
        p->lc_dep_styles = p->im_dep_styles ;
      } else {
        struct stat buf ;
        fstat(fd, &buf) ;
        size = ((buf.st_size+size)/size)*size ;
        if((int)(s = mmap(0,size ,PROT_READ,MAP_SHARED,fd,0)) == -1){
          p->lc_dep_styles = StrToIMStyles(buf.st_size, s);
        }
        close(fd);
      }
#else
      p->lc_dep_styles = p->im_dep_styles ;
#endif
      p->best_style= BestStyle(p->lc_dep_styles, p->im_dep_styles,
                               supported_styles) ;
    }
  }
  return(p->best_style);
}

#define MAXSTYLES 20
/*
 * This structure maps the input style name that a user or a system 
 * specifiy to the XIMStyle value.
 */
static struct {
      XIMStyle style ;
      char *name ;
      int namelen ;
  } XIMStyleRec[] = {
  XIMPreeditCallbacks|XIMStatusArea,"XIMPreeditCallbacks|XIMStatusArea",33,
  XIMPreeditCallbacks|XIMStatusNothing,
  "XIMPreeditCallbacks|XIMStatusNothing",36,
  XIMPreeditCallbacks|XIMStatusNone,"XIMPreeditCallbacks|XIMStatusNone",33,
  XIMPreeditCallbacks|XIMStatusCallbacks,
  "XIMPreeditCallbacks|XIMStatusCallbacks",38,
  XIMPreeditPosition|XIMStatusArea,"XIMPreeditPosition|XIMStatusArea",32,
  XIMPreeditPosition|XIMStatusNothing,
  "XIMPreeditPosition|XIMStatusNothing",35,
  XIMPreeditPosition|XIMStatusNone,"XIMPreeditPosition|XIMStatusNone",32,
  XIMPreeditPosition|XIMStatusCallbacks,
  "XIMPreeditPosition|XIMStatusCallbacks",37,
  XIMPreeditArea|XIMStatusArea,"XIMPreeditArea|XIMStatusArea",28,
  XIMPreeditArea|XIMStatusNothing,"XIMPreeditArea|XIMStatusNothing",31,
  XIMPreeditArea|XIMStatusNone,"XIMPreeditArea|XIMStatusNone",28,
  XIMPreeditArea|XIMStatusCallbacks,"XIMPreeditArea|XIMStatusCallbacks",33,
  XIMPreeditNothing|XIMStatusArea,"XIMPreeditNothing|XIMStatusArea",31,
  XIMPreeditNothing|XIMStatusNothing,"XIMPreeditNothing|XIMStatusNothing",34,
  XIMPreeditNothing|XIMStatusNone,"XIMPreeditNothing|XIMStatusNone",31,
  XIMPreeditNothing|XIMStatusCallbacks,
  "XIMPreeditNothing|XIMStatusCallbacks",36,
  XIMPreeditNone|XIMStatusArea,"XIMPreeditNone|XIMStatusArea",28,
  XIMPreeditNone|XIMStatusNothing,"XIMPreeditNone|XIMStatusNothing",31,
  XIMPreeditNone|XIMStatusNone,"XIMPreeditNone|XIMStatusNone",28,
  XIMPreeditNone|XIMStatusCallbacks,"XIMPreeditNone|XIMStatusCallbacks",33,
} ;

/*
 * Convert the string containing the input style priority list
 * in the format like 
 *
 XIMPreeditPosition|XIMStatusArea
 XIMPreeditPosition|XIMStatusNothing
 XIMPreeditArea|XIMStatusArea
 XIMPreeditArea|XIMStatusNothing
 XIMPreeditNothing|XIMStatusArea
 XIMPreeditNothing|XIMStatusNothing
 *
 * to the XIMStyles structure.
 * 
 */
static XIMStyles * StrToIMStyles(int size, char *s)
{
  XIMStyles *p = (XIMStyles *)calloc(sizeof(XIMStyles),1);
  register int i ;
  register char *c ;
  register char *end = s + size ;
  /*
   * assuming at least following 4 styles (Nothing * None)
   * are supported.
   */
  p->count_styles = 4 ;
  p->supported_styles = (XIMStyle *)calloc(sizeof(XIMStyle),
                        MAXSTYLES+4);
  p->supported_styles[0] = XIMPreeditNothing|XIMStatusNothing ;
  p->supported_styles[1] = XIMPreeditNothing|XIMStatusNone ;
  p->supported_styles[2] = XIMPreeditNone|XIMStatusNothing ;
  p->supported_styles[3] = XIMPreeditNone|XIMStatusNone ;
  /*
   * The following routine assumes that the style name listed in
   * InputStyle file is identical with the programatic name of style.
   * for example, "XIMPreeditPosition|XIMStatusArea" means the 
   * XIMPreeditPosition|XIMStatusArea value is specified. If the
   * style name is changed, such as "OverTheSpot|imDisplaysInClient",
   * the parsing logic below should be modified as well.
   */
  for(c = strchr(s, 'X') ; c < end ;){
    for( i=0; i<MAXSTYLES; i++ ){
      if(!strncmp(c, XIMStyleRec[i].name, XIMStyleRec[i].namelen)){
        p->count_styles++ ;
        p->supported_styles[p->count_styles-1] = XIMStyleRec[i].style ;
        c += XIMStyleRec[i].namelen ;
        break ;
      }
    }
    if(!(c = strchr(c, 'X'))){
      break ;
    }
  }
  return(p);
}

/* Choose the best style */
static XIMStyle BestStyle( XIMStyles *prio_order,
                           XIMStyles *im_support, XIMStyle *client_support)
{
  register unsigned short i, j, k ;
  register XIMStyle best_style ;

  for( i = 0 ; i<prio_order->count_styles ; i++ ) {
    best_style = prio_order->supported_styles[i];
    for( j = 0 ; j<im_support->count_styles ; j++ ) {
      if(im_support->supported_styles[i] == best_style ){
        for (k = 0 ; k<im_support->count_styles ; k++){
          if(client_support[i] == best_style ){
            return(best_style);
          }
        }
      }
    }
  }
  return(DEFAULTStyle);
}

/**
 *  name : XNPreEditAttributes or XNStatusAttributes
 *  area ; the constraints on the area
 **/
static void GetPreferredGeometry( XIC ic, char *name, XRectangle *area)
{
  XVaNestedList list;
  list = XVaCreateNestedList(0, XNAreaNeeded, area, NULL);
  /* set the constraints */
  XSetICValues(ic, name, list, NULL);
  /* Now query the preferred size */
  XGetICValues(ic, name, list, NULL);
  XFree(list);
}

static void SetGeometry( XIC ic, char *name, XRectangle *area)
{
  XVaNestedList list;
  list = XVaCreateNestedList(0, XNArea, area, NULL);
  XSetICValues(ic, name, list, NULL);
  XFree(list);
}
